#include <Leadwerks.h>
#include <wchar.h>
#include <codecvt>

namespace Leadwerks
{
    const int Font::Pixel = 0;
    const int Font::Smooth = 1;
    const int Font::LCD = 2;
    const int Font::English = 0;

	std::string WideStringToString(const std::wstring & str)
	{
		std::wstring_convert<std::codecvt_utf8<wchar_t>> myconv;
		return myconv.to_bytes(str);
	}

	std::wstring StringToWideString(const std::string & str)
	{
		std::wstring_convert<std::codecvt_utf8<wchar_t>> convert;
		return convert.from_bytes(str);
	}

	Font::Font() : faceinitialized(false), textureatlasx(0), textureatlasy(0), textureatlasmaxy(0), vertical(false), texture(NULL), style(0)
	{
		/*if (!FreeTypeInitialized__)
		{
        	if (FT_Init_FreeType( &freetypelibrary ))
			{
				System::Print("Warning: Error initializing FreeType");
			}
			else
			{
#ifdef _WIN32
				//FT_Library_SetLcdFilter(freetypelibrary, FT_LCD_FILTER_DEFAULT);
#endif
				FreeTypeInitialized__=true;
			}
		}*/
	}

	Font::~Font()
	{
		Reset();
		if (path!="")
		{
			System::Print("Deleting font \""+path+"\"");
			path += + "|" + String(size) + "|" + String(style) + "|" + String(family);
		}
	}

	void Font::Reset()
	{
		Character* c;
		//map<int,Character*>::reverse_iterator rit;
		map<int,Character*>::iterator it;

		faceinitialized = false;
		if (texture)
		{
			texture->Release();
			texture=NULL;
		}
		for (it=character.begin(); it != character.end(); it++)
		{
			c = it->second;
			delete c;
		}
		character.clear();
	}

	std::string Font::GetClassName()
	{
		return "Font";
	}

	Asset* Font::Copy()
	{
		map<int,Character*>::reverse_iterator rit;
		Font* font = new Font;

		font->driver = driver;
		font->height = height;
		font->faceinitialized = faceinitialized;
		font->textureatlasx = textureatlasx;
		font->textureatlasy = textureatlasy;
		font->textureatlasmaxy = textureatlasmaxy;
		font->style = style;
		font->size = size;
		font->vertical = vertical;
		font->texture = (Texture*)texture->Copy();
		for ( rit=character.rbegin(); rit != character.rend(); rit++)
		{
			font->character[rit->first] = (rit->second)->Copy();
		}
		return font;
	}

	bool Font::Reload(Stream* stream, const int flags)
	{
		unsigned int i;
		Font* font;
		Character* ch;
        int version;
        int facecount;
        int facesize;
        int charcount;
        int pos;
        int texwidth;
        int texheight;
        int compressed;
        Bank* decompressedbank=NULL;
        int n,c,texformat,bpp;
        Bank* pixels = NULL;
		std::wstring familychars;
		Bank* data;
		int fsize = stream->GetSize();
		FT_Library freetypelibrary;

		if (FT_Init_FreeType( &freetypelibrary )!=0)
		{
			System::Print("Error: Failed to initialize FreeType library.");
			return false;
		}

		data = Bank::Create(fsize);

		stream->Read(data->buf,fsize);

		if (FT_New_Memory_Face( freetypelibrary, (const FT_Byte*)data->buf, fsize, 0, &face ))
		{
			System::Print("Error: Failed to load font.");
			FT_Done_Face(face);
			delete data;
			FT_Done_FreeType(freetypelibrary);
			return false;
		}
		else
		{
			faceinitialized=true;
		}

		if (FT_Set_Char_Size(face, 0, size * 64, 0, 96))
		{
			System::Print("Error: Error setting font size.");
			FT_Done_Face(face);
			delete data;
			FT_Done_FreeType(freetypelibrary);
			return false;
		}

		//FT_Library_SetLcdFilter(freetypelibrary, FT_LCD_FILTER_LIGHT);

        if (family==Font::English)
        {
			familychars = L"abcdefghijklmnopqrstuvwxyz ŨABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-=!@#$%^&*()_+[]\\{}|;':\",./<>?	`~";
		}
		else
        {
			System::Print("Error: Failed to load font.");
            FT_Done_Face(face);
			delete data;
			FT_Done_FreeType(freetypelibrary);
			return false;
		}

		//int familycharsLenght = wcslen(familychars);
		for (i = 0; i < familychars.length(); i++)
		{
			ch = LoadChar(familychars[i]);
		}

		if (FT_HAS_VERTICAL(face)==0)
		{
			vertical = false;
		}
		else
		{
			vertical = true;
		}

		if (!BuildAtlas())
		{
			System::Print("Error: Failed to load font.");
			FT_Done_Face(face);
			delete data;
			Reset();
			FT_Done_FreeType(freetypelibrary);
			return false;
		}

		FT_Done_Face(face);
		delete data;
		faceinitialized = false;
		FT_Done_FreeType(freetypelibrary);

		return true;
	}

	bool Font::Reload(const int flags)
	{
		if (path=="") return false;

		Reset();

		std::vector<std::string> sarr = String::Split(path, "|");

		if (fileid == 0)
		{
			System::Print("Loading font \"" + path + "\"...");
		}
		else
		{
			System::Print("Loading font \"" + path + "\""+String(fileid)+"...");
		}

		//FT_ULong charcode;
		//FT_UInt gindex;
		unsigned int i;
		Font* font;
		Character* ch;
        int version;
        int facecount;
        int facesize;
        int charcount;
        int pos;
        int texwidth;
        int texheight;
        int compressed;
        Bank* decompressedbank=NULL;
        int n,c,texformat,bpp;
        Bank* pixels = NULL;
        Stream* stream = NULL;
		bool result;

		//std::vector<std::string> sarr = String::Split(path,"|");

        //System::Print("Loading font \""+sarr[0]+"\" ("+sarr[1]+", "+sarr[2]+", "+sarr[2]+")...");

		/*if (sarr.size()>0)
		{
			stream = FileSystem::ReadFile(sarr[0]);
			if (stream==NULL)
			{
				System::Print("Warning: Failed to read file \""+sarr[0]+"\".");
				return false;
			}
		}
		else
		{
			System::Print("Warning: Failed to load font.");
			return false;
		}*/

		//std::string rpath = FileSystem::RealPath(path,fileid);
		//stream = FileSystem::ReadFile(rpath);
		//if (!stream) return false;
		//bool result = Reload(stream,flags);
		//delete stream;
		//return result;

		wchar_t* familychars;
		std::string key = Leadwerks::String::XOR(Leadwerks::String(Leadwerks::Steamworks::GetUserID()) + Leadwerks::String(fileid), Leadwerks::String::Chr(43) + Leadwerks::String::Chr(13) + Leadwerks::String::Chr(105) + Leadwerks::String::Chr(92) + Leadwerks::String::Chr(137) + Leadwerks::String::Chr(68));
        stream = FileSystem::ReadFile(path,fileid,key);
		if (!stream)
        {
            System::Print("Failed to read file \""+path+"\".");
            return false;
        }
		result = Reload(stream,flags);
		delete stream;

		return result;

		/*
		int fsize = stream->GetSize();
		Bank* bank = Bank::Create(fsize);
		stream->Read(bank->buf,fsize);
		delete stream;
		//Bank* bank = Bank::Load(rpath);
		if (FT_New_Memory_Face( freetypelibrary, (const FT_Byte*)bank->buf, bank->GetSize(), 0, &face ))
		//if (FT_New_Face( freetypelibrary, rpath.c_str(), 0, &face ))
		{
			System::Print("Failed to load font \""+rpath+"\"");
			return false;
		}
		else
		{
			faceinitialized=true;
		}

		if (FT_Set_Char_Size(face, 0, size * 64, 0, 96))
		{
			System::Print("Error setting font size.");
			return false;
		}

		//FT_Library_SetLcdFilter(freetypelibrary, FT_LCD_FILTER_LIGHT);

		switch (family)
		{
		case FONT_ENGLISH:
			familychars = "abcdefghijklmnopqrstuvwxyz ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-=!@#$%^&*()_+[]\\{}|;':\",./<>?	`~";
			break;
		default:
			return false;
		}

		//Load characters and save info to text file
		for (i=0; i<familychars.length(); i++)
		{
			ch = LoadChar(familychars[i]);
		}

		if (FT_HAS_VERTICAL(face)==0)
		{
			vertical = false;
		}
		else
		{
			vertical = true;
		}

		if (!BuildAtlas())
		{
			Reset();
			return false;
		}
		FT_Done_Face(face);
		faceinitialized = false;

		return true;*/
	}

	int Font::GetTextWidth(const std::string& text)
	{
		unsigned int i;
		int lx=0;
		int dx=0;
		std::wstring temp = StringToWideString(text);
		for (i=0; i< temp.length(); i++)
		{
			Character* c = character[temp[i]];
			if (c!=NULL)
			{
				//if (dx+c->x<=lx) dx = lx-c->x+1;
				lx = dx + c->width-1;
				dx += c->spacing;
			}
		}
		return dx;
	}

	Character* Font::LoadChar(const int c, const bool force)
	{
		//Bank* pixels;
		int i;
		int width;
		int height;
		int pitch;
		int tw;
		int th;
		int x,y;
		unsigned char r,g,b;
		Character* ch = NULL;
		int loadflags = FT_LOAD_DEFAULT;
        int bpp = 1;

        //Load and render the glyph
		if (style==Font::LCD) loadflags = FT_LOAD_TARGET_LCD;
        FT_UInt glyph_index = FT_Get_Char_Index(face,c);
		if (glyph_index==0) return NULL;
		if (FT_Load_Glyph (face, glyph_index, loadflags)) return NULL;
		if (face->glyph->format != FT_GLYPH_FORMAT_BITMAP)
		{
			if (style==Font::Pixel)
            {
				if (FT_Render_Glyph (face->glyph, FT_RENDER_MODE_MONO)) return NULL;
			}
            else if (style==Font::Smooth)
            {
                if (FT_Render_Glyph (face->glyph, FT_RENDER_MODE_NORMAL)) return NULL;
            }
			else if (style==Font::LCD)
            {
				if (FT_Render_Glyph (face->glyph, FT_RENDER_MODE_LCD)) return NULL;
			}
		}

        if (style==Font::LCD) bpp = 3;

		width  = face->glyph->bitmap.width;
		height = face->glyph->bitmap.rows;
		pitch = face->glyph->bitmap.pitch;

		if (vertical)
		{
			if (style==Font::LCD)
			{
				th = height/3;
			}
			else
			{
				th = height;
			}
			tw = width;
		}
		else
		{
			if (style==Font::LCD)
			{
				tw = width/3;
			}
			else
			{
				tw = width;
			}
			th = height;
		}

		if (c==77) this->height=th;

		//Set the character parameters
		ch = new Character;
		character[c] = ch;
		ch->x = face->glyph->metrics.horiBearingX/64;
		ch->y = face->glyph->metrics.horiBearingY/64;
		ch->width = tw;
		ch->height = th;
		ch->id = c;

		if (vertical)
		{
			ch->spacing = face->glyph->metrics.vertAdvance/64.0;
		}
		else
		{
			ch->spacing = face->glyph->metrics.horiAdvance/64.0;
		}

		//Copy the glyph image to the character data bank
		ch->data = Bank::Create( tw * th * bpp );

		//Copy the glyph bitmap to the character data
		for (x=0; x<tw; x++)
		{
			for (y=0; y<th; y++)
			{
				if (style==Font::Pixel)
                {
					memcpy(&r, face->glyph->bitmap.buffer + int(floor(float(y*pitch + x / 8))), 1);
					i = 7-int(Math::Mod(x,8));
					if (int(pow(2.0,i)) & r)
					{
						r = 255;
					}
					else
					{
						r = 0;
					}
					g=r;
					b=r;
                    ch->data->PokeByte((x+y*tw)*1+0,r);
                }
				else if (style==Font::Smooth)
                {
					memcpy(&r, face->glyph->bitmap.buffer + y*pitch + x, 1);
					g=r;
					b=r;
                    ch->data->PokeByte((x+y*tw)*1+0,r);
				}
                else if (style==Font::LCD)
                {
                    memcpy(&r, face->glyph->bitmap.buffer + y*pitch + x*3 + 0, 1);
                    memcpy(&g, face->glyph->bitmap.buffer + y*pitch + x*3 + 1, 1);
                    memcpy(&b, face->glyph->bitmap.buffer + y*pitch + x*3 + 2, 1);
                    ch->data->PokeByte((x+y*tw)*3+0,255-r);
                    ch->data->PokeByte((x+y*tw)*3+1,255-g);
                    ch->data->PokeByte((x+y*tw)*3+2,255-b);
				}
			}
		}
		return ch;
	}

	bool Font::BuildAtlas()
	{
		map<int,Character*>::reverse_iterator rit;
		int tw,th,x,y,r,g,b;//,a;
		int atlassize = 128;
		bool finished = false;
		Character* c;

		//Calculate atlas size.  This loop will keep repeating until it finds the minimum texture
		//size within which all characters will fit.
		for(;;)
		{
			//Reset values
			textureatlasmaxy = 0;
			textureatlasx = 0;
			textureatlasy = 0;
			finished = true;

			for ( rit=character.rbegin() ; rit != character.rend(); rit++)
			{
				c = rit->second;
				tw = c->width;
				th = c->height;

				//Move to the next vertical line if we run out of horizontal space
				if (textureatlasx+tw>atlassize)
				{
					textureatlasx = 0;
					textureatlasy = textureatlasmaxy+1;
					textureatlasmaxy = textureatlasy;
				}

				//If there's not enough vertical space, resize texture atlas and start over
				if (textureatlasy+th>atlassize)
				{
					atlassize *= 2;
					finished = false;
					break;
				}

				c->atlasx = textureatlasx;
				c->atlasy = textureatlasy;
				c->texcoords.x = float(textureatlasx)/atlassize;
				c->texcoords.y = float(textureatlasy)/atlassize;
				c->texcoords.z = float(textureatlasx+c->width)/atlassize;
				c->texcoords.w = float(textureatlasy+c->height)/atlassize;

				textureatlasmaxy = max(textureatlasmaxy,textureatlasy+th);
				textureatlasx += tw+1;
			}
			if (finished) break;
		}

		//Create texture
		if (texture) texture->Release();
		texture = Texture::Create(atlassize,atlassize,Texture::RGBA);
        Debug::Assert(texture);
		if (style==Font::Smooth)
		{
			texture->SetFilter(Texture::Smooth);
		}
		else
		{
			texture->SetFilter(Texture::Pixel);
		}

		Bank* pixels = Bank::Create(atlassize*atlassize*4);
		pixels->Clear();

		for (rit=character.rbegin(); rit != character.rend(); rit++)
		{
			c = rit->second;
			for (x=0; x<c->width; x++)
			{
				for (y=0; y<c->height; y++)
				{
                    if (style==Font::LCD)
                    {
                        r = rit->second->data->PeekByte((y*c->width+x)*3+0);
                        g = rit->second->data->PeekByte((y*c->width+x)*3+1);
                        b = rit->second->data->PeekByte((y*c->width+x)*3+2);
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+0,r);
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+1,g);
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+2,b);
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+3,255);
                    }
                    else
                    {
                        r = rit->second->data->PeekByte((y*c->width+x)*1+0);
                        g=r;
                        b=r;
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+0,255);
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+1,255);
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+2,255);
                        pixels->PokeByte(((x + c->atlasx)+(y + c->atlasy)*atlassize)*4+3,r);
                    }
				}
			}
			//delete c->data;
			//c->data = NULL;
			texture->SetPixels(pixels->buf,0,0,0);
		}
		delete pixels;
		/*texture->Lock();
		for (rit=character.rbegin(); rit != character.rend(); rit++)
		{
			c = rit->second;
			for (x=0; x<c->width; x++)
			{
				for (y=0; y<c->height; y++)
				{
					r = rit->second->data->PeekByte((y*c->width+x)*3+0);
					g = rit->second->data->PeekByte((y*c->width+x)*3+1);
					b = rit->second->data->PeekByte((y*c->width+x)*3+2);
					texture->WritePixel( x + c->atlasx, y + c->atlasy, r, g, b );
				}
			}
			delete c->data;
			c->data = NULL;
		}
		texture->Unlock();
		*/
		return true;
	}

	void Font::DrawText(const std::wstring text, const float x, const float y, const float kerning)
	{
		unsigned int i;
		float dx = 0;
		float dy = 0;
		int lx = 0;
		int mode = 0;
		bool unloadshader = false;
		float kerning_ = kerning;

		if (kerning_ == 0.0) kerning_ = 1;
		texture->Bind();

		if (style == Font::LCD) mode = 1;
		((GraphicsDriver*)driver)->BeginChars(x, y, mode);

		//Draw each character
		for (i = 0; i < text.length(); i++)
		{
			Character* c = character[text[i]];
			if (c != NULL)
			{
				if (vertical)
				{
					((GraphicsDriver*)driver)->DrawChar(height + x - c->x, dy + c->y, c->width, c->height, c->texcoords);
					dy += ((float)c->spacing) * kerning_;
				}
				else
				{
					((GraphicsDriver*)driver)->DrawChar(dx + c->x, height - c->y, c->width, c->height, c->texcoords);
					lx = dx + c->width - 1;
					dx += ((float)c->spacing) * kerning_;
				}
			}
		}
		((GraphicsDriver*)driver)->EndChars();
	}

	void Font::DrawText(const std::string& text, const float x, const float y, const float kerning)
	{
		unsigned int i;
		float dx = 0;
		float dy = 0;
		int lx = 0;
		int mode=0;
		bool unloadshader = false;
		float kerning_ = kerning;

        if (kerning_==0.0) kerning_=1;

		texture->Bind();

		//Mat4 mat = driver->drawmatrix;
		//Mat4 temp = driver->drawmatrix;
		//mat.t.x += x;
		//mat[3].y += y;
		//driver->drawmatrix = mat;

		//if (shader!=NULL) shader->SetMat4("drawmatrix",mat);

		if (style==Font::LCD) mode=1;
        ((GraphicsDriver*)driver)->BeginChars(x,y,mode);

		//Draw each character
		//wstring test1 = wstring(text.begin(), text.end());
		//wchar_t* test2 = L"Hello";

		std::wstring temp = StringToWideString(text);
		for (i=0; i< temp.length(); i++)
		{
            Character* c = character[temp[i]];
			if (c!=NULL)
			{
            if (vertical)
				{
					((GraphicsDriver*)driver)->DrawChar(height+x-c->x,dy+c->y,c->width,c->height,c->texcoords);
					dy += ((float)c->spacing) * kerning_;
				}
				else
				{
					//Print(c->width);
					//Print(c->height);
					//Print(c->spacing);
                    //if (dx+c->x<=lx) dx = lx-c->x+1;
					((GraphicsDriver*)driver)->DrawChar(dx+c->x,height-c->y,c->width,c->height,c->texcoords);
					lx = dx + c->width-1;
					dx += ((float)c->spacing) * kerning_;
				}
			}
		}
        ((GraphicsDriver*)driver)->EndChars();

		//driver->drawmatrix = temp;
		//if (unloadshader) driver->SetShader(NULL);
	}

    int Font::GetHeight()
	{
		return height;
	}

    Texture* Font::GetTexture()
	{
        //if (texture) texture->AddRef();
		return texture;
	}

    Font* Font::Load(Stream* stream, const int size, const int style, const int family, const int flags)
	{
		GraphicsDriver* driver = GraphicsDriver::GetCurrent();
		Font* font = new Font;
		font->driver = driver;
		font->size = size;
		font->style = style;
		font->family = family;
		if (font->Reload(stream,flags))
		{
            return font;
		}
		else
		{
            font--;
			return NULL;
		}
	}

	Font* Font::Load(const std::string& path, const int size, const int style, const int family, const int flags, const uint64_t fileid)
	{
		GraphicsDriver* driver = GraphicsDriver::GetCurrent();
		Font* font = NULL;
		Asset* asset = NULL;
        std::string lpath;
		std::string rpath;
		if (fileid == 0)
		{
			rpath = FileSystem::RealPath(path);
#ifdef __linux__
			if (FileSystem::GetFileType(rpath)==0)
			{
				rpath = FileSystem::CasedPath(rpath);
				if (rpath=="")
				{
					System::Print("Error: Failed to load font \"" + path + "\"");
					return false;
				}
			}
#endif
		}
		else
		{
			rpath = path;
		}
		std::string storagepath = rpath + "|" + String(size) + "|" + String(style) + "|" + String(family);
#ifdef __linux__
		lpath=storagepath;
#else
		lpath = String::Lower(storagepath);
#endif

		if (!(Asset::Unmanaged & flags))
		{
			asset = driver->FindAsset(lpath,fileid);
			if (asset)
            {
                asset->AddRef();
                return (Font*)asset;
            }
		}

		font = new Font;
		font->fileid = fileid;
		font->driver = driver;
		font->size = size;
		font->style = style;
		font->family = family;
		font->path = rpath;
		font->storagepath = storagepath;

		if (font->Reload(flags) || (Asset::CreateNew & flags))
		{
			if (!(Asset::Unmanaged & flags))
			{
				font->driver->workshopfiles[fileid][storagepath] = font;
                //font->driver->loadedassets[lpath]=font;
			}
            return font;
		}
		else
		{
			System::Print("Error: Failed to load font \"" + font->path + "\"");
			font->path="";
            font->Release();
			return NULL;
		}
	}

}
